import typing as ta

from .types import Annotated
from .types import Any
from .types import Generic
from .types import NewType
from .types import Type
from .types import Union
from .types import type_


def strip_objs(ty: Type) -> Type:
    if isinstance(ty, (type, ta.TypeVar, NewType, Any)):
        return ty

    if isinstance(ty, Union):
        return Union(frozenset(map(strip_objs, ty.args)))

    if isinstance(ty, Generic):
        return Generic(ty.cls, tuple(map(strip_objs, ty.args)), ty.params, None)

    if isinstance(ty, Annotated):
        return Annotated(strip_objs(ty.ty), ty.md, None)

    raise TypeError(ty)


def strip_annotations(ty: Type) -> Type:
    if isinstance(ty, (type, ta.TypeVar, NewType, Any)):
        return ty

    if isinstance(ty, Union):
        return Union(frozenset(map(strip_annotations, ty.args)))

    if isinstance(ty, Generic):
        return Generic(ty.cls, tuple(map(strip_annotations, ty.args)), ty.params, ty.obj)

    if isinstance(ty, Annotated):
        return strip_annotations(ty.ty)

    raise TypeError(ty)


def types_equivalent(l: Type, r: Type) -> bool:
    if isinstance(l, Generic) and isinstance(r, Generic):
        return l.cls == r.cls and l.args == r.args

    return l == r


def get_underlying(nt: NewType) -> Type:
    return type_(nt.obj.__supertype__)  # noqa


def get_concrete_type(ty: Type) -> type | None:
    if isinstance(ty, type):
        return ty

    if isinstance(ty, Generic):
        return ty.cls

    if isinstance(ty, NewType):
        return get_concrete_type(get_underlying(ty))

    if isinstance(ty, (Union, ta.TypeVar, Any)):
        return None

    raise TypeError(ty)


def to_annotation(ty: Type) -> ta.Any:
    if isinstance(ty, Generic):
        return ty.obj if ty.obj is not None else ty.cls

    if isinstance(ty, Union):
        return ta.Union[*tuple(to_annotation(e) for e in ty.args)]

    if isinstance(ty, (type, ta.TypeVar, NewType)):
        return ty

    if isinstance(ty, Any):
        return ta.Any

    raise TypeError(ty)



import typing as ta

from .. import lang
from .keys import Key


if ta.TYPE_CHECKING:
    from .impl import inspect as _inspect
else:
    _inspect = lang.proxy_import('.impl.inspect', __package__)


T = ta.TypeVar('T')


class Kwarg(ta.NamedTuple):
    name: str
    key: Key
    has_default: bool


class KwargsTarget(ta.NamedTuple):
    obj: ta.Any
    kwargs: ta.Sequence[Kwarg]


def tag(obj: T, **kwargs: ta.Any) -> T:
    return _inspect.tag(obj, **kwargs)


def build_kwargs_target(obj: ta.Any, **kwargs: ta.Any) -> KwargsTarget:
    return _inspect.build_kwargs_target(obj, **kwargs)



import dataclasses as dc

from .fields import field_type
from .internals import FIELDS_ATTR
from .internals import FieldType
from .internals import is_dataclass_instance
from .processing import Processor
from .utils import set_new_attribute


MISSING = dc.MISSING


def replace(obj, /, **changes):  # noqa
    if not is_dataclass_instance(obj):
        raise TypeError('replace() should be called on dataclass instances')
    return _replace(obj, **changes)


def _replace(obj, /, **changes):
    for f in getattr(obj, FIELDS_ATTR).values():
        if (ft := field_type(f)) is FieldType.CLASS:
            continue

        if not f.init:
            if f.name in changes:
                raise TypeError(f'field {f.name} is declared with init=False, it cannot be specified with replace()')
            continue

        if f.name not in changes:
            if ft is FieldType.INIT and f.default is MISSING:
                raise TypeError(f'InitVar {f.name!r} must be specified with replace()')
            changes[f.name] = getattr(obj, f.name)

    return obj.__class__(**changes)


class ReplaceProcessor(Processor):
    def _process(self) -> None:
        set_new_attribute(self._cls, '__replace__', _replace)



# ruff: noqa: UP006 UP007
import contextlib
import logging
import os
import shlex
import subprocess
import sys
import time
import typing as ta

from .logs import log
from .runtime import is_debugger_attached


T = ta.TypeVar('T')
SubprocessChannelOption = ta.Literal['pipe', 'stdout', 'devnull']


##


SUBPROCESS_CHANNEL_OPTION_VALUES: ta.Mapping[SubprocessChannelOption, int] = {
    'pipe': subprocess.PIPE,
    'stdout': subprocess.STDOUT,
    'devnull': subprocess.DEVNULL,
}


##


_SUBPROCESS_SHELL_WRAP_EXECS = False


def subprocess_shell_wrap_exec(*args: str) -> ta.Tuple[str, ...]:
    return ('sh', '-c', ' '.join(map(shlex.quote, args)))


def subprocess_maybe_shell_wrap_exec(*args: str) -> ta.Tuple[str, ...]:
    if _SUBPROCESS_SHELL_WRAP_EXECS or is_debugger_attached():
        return subprocess_shell_wrap_exec(*args)
    else:
        return args


def prepare_subprocess_invocation(
        *args: str,
        env: ta.Optional[ta.Mapping[str, ta.Any]] = None,
        extra_env: ta.Optional[ta.Mapping[str, ta.Any]] = None,
        quiet: bool = False,
        shell: bool = False,
        **kwargs: ta.Any,
) -> ta.Tuple[ta.Tuple[ta.Any, ...], ta.Dict[str, ta.Any]]:
    log.debug('prepare_subprocess_invocation: args=%r', args)
    if extra_env:
        log.debug('prepare_subprocess_invocation: extra_env=%r', extra_env)

    if extra_env:
        env = {**(env if env is not None else os.environ), **extra_env}

    if quiet and 'stderr' not in kwargs:
        if not log.isEnabledFor(logging.DEBUG):
            kwargs['stderr'] = subprocess.DEVNULL

    if not shell:
        args = subprocess_maybe_shell_wrap_exec(*args)

    return args, dict(
        env=env,
        shell=shell,
        **kwargs,
    )


##


@contextlib.contextmanager
def subprocess_common_context(*args: ta.Any, **kwargs: ta.Any) -> ta.Iterator[None]:
    start_time = time.time()
    try:
        log.debug('subprocess_common_context.try: args=%r', args)
        yield

    except Exception as exc:  # noqa
        log.debug('subprocess_common_context.except: exc=%r', exc)
        raise

    finally:
        end_time = time.time()
        elapsed_s = end_time - start_time
        log.debug('subprocess_common_context.finally: elapsed_s=%f args=%r', elapsed_s, args)


##


def subprocess_check_call(
        *args: str,
        stdout: ta.Any = sys.stderr,
        **kwargs: ta.Any,
) -> None:
    args, kwargs = prepare_subprocess_invocation(*args, stdout=stdout, **kwargs)
    with subprocess_common_context(*args, **kwargs):
        return subprocess.check_call(args, **kwargs)  # type: ignore


def subprocess_check_output(
        *args: str,
        **kwargs: ta.Any,
) -> bytes:
    args, kwargs = prepare_subprocess_invocation(*args, **kwargs)
    with subprocess_common_context(*args, **kwargs):
        return subprocess.check_output(args, **kwargs)


def subprocess_check_output_str(*args: str, **kwargs: ta.Any) -> str:
    return subprocess_check_output(*args, **kwargs).decode().strip()


##


DEFAULT_SUBPROCESS_TRY_EXCEPTIONS: ta.Tuple[ta.Type[Exception], ...] = (
    FileNotFoundError,
    subprocess.CalledProcessError,
)


def _subprocess_try_run(
        fn: ta.Callable[..., T],
        *args: ta.Any,
        try_exceptions: ta.Tuple[ta.Type[Exception], ...] = DEFAULT_SUBPROCESS_TRY_EXCEPTIONS,
        **kwargs: ta.Any,
) -> ta.Union[T, Exception]:
    try:
        return fn(*args, **kwargs)
    except try_exceptions as e:  # noqa
        if log.isEnabledFor(logging.DEBUG):
            log.exception('command failed')
        return e


def subprocess_try_call(
        *args: str,
        try_exceptions: ta.Tuple[ta.Type[Exception], ...] = DEFAULT_SUBPROCESS_TRY_EXCEPTIONS,
        **kwargs: ta.Any,
) -> bool:
    if isinstance(_subprocess_try_run(
            subprocess_check_call,
            *args,
            try_exceptions=try_exceptions,
            **kwargs,
    ), Exception):
        return False
    else:
        return True


def subprocess_try_output(
        *args: str,
        try_exceptions: ta.Tuple[ta.Type[Exception], ...] = DEFAULT_SUBPROCESS_TRY_EXCEPTIONS,
        **kwargs: ta.Any,
) -> ta.Optional[bytes]:
    if isinstance(ret := _subprocess_try_run(
            subprocess_check_output,
            *args,
            try_exceptions=try_exceptions,
            **kwargs,
    ), Exception):
        return None
    else:
        return ret


def subprocess_try_output_str(*args: str, **kwargs: ta.Any) -> ta.Optional[str]:
    out = subprocess_try_output(*args, **kwargs)
    return out.decode().strip() if out is not None else None


##


def subprocess_close(
        proc: subprocess.Popen,
        timeout: ta.Optional[float] = None,
) -> None:
    # TODO: terminate, sleep, kill
    if proc.stdout:
        proc.stdout.close()
    if proc.stderr:
        proc.stderr.close()
    if proc.stdin:
        proc.stdin.close()

    proc.wait(timeout)



"""
TODO:
 - required_keys
"""
import abc
import typing as ta

from ... import dataclasses as dc
from ... import lang
from ... import reflect as rfl
from ..injector import Injector
from ..inspect import KwargsTarget
from ..providers import ConstProvider
from ..providers import CtorProvider
from ..providers import FnProvider
from ..providers import LinkProvider
from ..providers import Provider
from .inspect import build_kwargs_target


##


class ProviderImpl(lang.Abstract):
    @property
    @abc.abstractmethod
    def providers(self) -> ta.Iterable[Provider]:
        raise NotImplementedError

    @abc.abstractmethod
    def provide(self, injector: Injector) -> ta.Any:
        raise NotImplementedError


##


@dc.dataclass(frozen=True, eq=False)
class InternalProvider(Provider):
    impl: ProviderImpl

    def provided_ty(self) -> rfl.Type | None:
        raise TypeError


##


@dc.dataclass(frozen=True, eq=False)
class CallableProviderImpl(ProviderImpl, lang.Final):
    p: FnProvider | CtorProvider
    kt: KwargsTarget

    @property
    def providers(self) -> ta.Iterable[Provider]:
        return (self.p,)

    def provide(self, injector: Injector) -> ta.Any:
        return injector.inject(self.kt)


##


@dc.dataclass(frozen=True, eq=False)
class ConstProviderImpl(ProviderImpl, lang.Final):
    p: ConstProvider

    @property
    def providers(self) -> ta.Iterable[Provider]:
        return (self.p,)

    def provide(self, injector: Injector) -> ta.Any:
        return self.p.v


##


@dc.dataclass(frozen=True, eq=False)
class LinkProviderImpl(ProviderImpl, lang.Final):
    p: LinkProvider

    @property
    def providers(self) -> ta.Iterable[Provider]:
        return (self.p,)

    def provide(self, injector: Injector) -> ta.Any:
        return injector.provide(self.p.k)


##


PROVIDER_IMPLS_BY_PROVIDER: dict[type[Provider], ta.Callable[..., ProviderImpl]] = {
    FnProvider: lambda p: CallableProviderImpl(p, build_kwargs_target(p.fn)),
    CtorProvider: lambda p: CallableProviderImpl(p, build_kwargs_target(p.ty)),
    ConstProvider: ConstProviderImpl,
    LinkProvider: LinkProviderImpl,
    InternalProvider: lambda p: p.impl,
}


def make_provider_impl(p: Provider) -> ProviderImpl:
    try:
        fac = PROVIDER_IMPLS_BY_PROVIDER[type(p)]
    except KeyError:
        pass
    else:
        return fac(p)

    raise TypeError(p)



import dataclasses as dc
import os
import subprocess

from .. import lang


@dc.dataclass(frozen=True)
class PsItem:
    pid: int
    ppid: int
    cmd: str


def get_ps_item(pid: int, timeout: lang.Timeout | None = None) -> PsItem:
    timeout = lang.timeout(timeout)
    out = subprocess.check_output(
        ['ps', '-o', 'pid=,ppid=,command=', str(int(pid))],
        timeout=timeout.or_(None),
    ).decode().strip()
    opid, _, rest = out.partition(' ')
    ppid, _, cmd = rest.strip().partition(' ')
    return PsItem(
        int(opid),
        int(ppid),
        cmd.strip(),
    )


def get_ps_lineage(pid: int, timeout: lang.Timeout | None = None) -> list[PsItem]:
    timeout = lang.timeout(timeout)
    ret: list[PsItem] = []
    while True:
        cur = get_ps_item(pid, timeout)
        if cur.ppid < 1:
            break
        ret.append(cur)
        pid = cur.ppid
    return ret


def _main() -> None:
    print(get_ps_lineage(os.getpid()))


if __name__ == '__main__':
    _main()



import abc
import enum
import typing as ta

from .. import check
from .. import lang


##


ParamKey: ta.TypeAlias = str | int


class ParamsPreparer(lang.Abstract):
    @abc.abstractmethod
    def add(self, k: ParamKey) -> str:
        raise NotImplementedError

    @abc.abstractmethod
    def prepare(self) -> lang.Args:
        raise NotImplementedError


class LinearParamsPreparer(ParamsPreparer):
    def __init__(self, placeholder: str) -> None:
        super().__init__()
        self._placeholder = placeholder
        self._args: list[ParamKey] = []

    def add(self, k: ParamKey) -> str:
        self._args.append(k)
        return self._placeholder

    def prepare(self) -> lang.Args:
        return lang.Args(*self._args)


class NumericParamsPreparer(ParamsPreparer):
    def __init__(self) -> None:
        super().__init__()
        self._args: list[ParamKey] = []
        self._str_by_key: dict[ParamKey, str] = {}
        self._pos_by_name: dict[str, int] = {}
        self._pos_by_id: dict[int, int] = {}

    def add(self, k: ParamKey) -> str:
        try:
            return self._str_by_key[k]
        except KeyError:
            pass

        pos = len(self._args)
        if isinstance(k, str):
            check.not_in(k, self._pos_by_name)
            self._pos_by_name[k] = pos
        elif isinstance(k, int):
            check.not_in(k, self._pos_by_id)
            self._pos_by_id[k] = pos
        else:
            raise TypeError(f'Invalid key type: {type(k)}')

        self._args.append(k)
        ret = self._str_by_key[k] = f':{pos + 1}'
        return ret

    def prepare(self) -> lang.Args:
        return lang.Args(*self._args)


class NamedParamsPreparer(ParamsPreparer):
    def __init__(
            self,
            render: ta.Callable[[str], str],
            *,
            anon: ta.Callable[[int], str] | None = None,
    ) -> None:
        super().__init__()
        self._render = render
        self._anon = anon if anon is not None else self.underscore_anon
        self._kwargs: dict[str, ParamKey] = {}
        self._str_by_key: dict[ParamKey, str] = {}
        self._kwargs_by_name: dict[str, str] = {}
        self._kwargs_by_id: dict[int, str] = {}

    @staticmethod
    def render_named(kwarg: str) -> str:
        return f':{kwarg}'

    @staticmethod
    def render_pyformat(kwarg: str) -> str:
        return f'%({kwarg})s'

    @staticmethod
    def underscore_anon(n: int) -> str:
        return f'_{n}'

    def add(self, k: ParamKey) -> str:
        try:
            return self._str_by_key[k]
        except KeyError:
            pass

        kwarg: str
        if isinstance(k, str):
            check.not_in(k, self._kwargs_by_name)
            kwarg = k  # TODO: collisions
            self._kwargs_by_name[k] = kwarg
        elif isinstance(k, int):
            check.not_in(k, self._kwargs_by_id)
            kwarg = self._anon(len(self._kwargs_by_id))  # TODO: collisions
            self._kwargs_by_id[k] = kwarg
        else:
            raise TypeError(f'Invalid key type: {type(k)}')

        self._kwargs[kwarg] = k
        ret = self._str_by_key[k] = self._render(kwarg)
        return ret

    def prepare(self) -> lang.Args:
        return lang.Args(**self._kwargs)


##


class ParamStyle(enum.Enum):
    QMARK = 'qmark'  # Question mark style, e.g. ...WHERE name=?
    FORMAT = 'format'  # ANSI C printf format codes, e.g. ...WHERE name=%s

    NUMERIC = 'numeric'  # Numeric, positional style, e.g. ...WHERE name=:1

    NAMED = 'named'  # Named style, e.g. ...WHERE name=:name
    PYFORMAT = 'pyformat'  # Python extended format codes, e.g. ...WHERE name=%(name)s


def make_params_preparer(style: ParamStyle) -> ParamsPreparer:
    if style is ParamStyle.QMARK:
        return LinearParamsPreparer('?')
    elif style is ParamStyle.FORMAT:
        return LinearParamsPreparer('%s')
    elif style is ParamStyle.NUMERIC:
        return NumericParamsPreparer()
    elif style is ParamStyle.NAMED:
        return NamedParamsPreparer(NamedParamsPreparer.render_named)
    elif style is ParamStyle.PYFORMAT:
        return NamedParamsPreparer(NamedParamsPreparer.render_pyformat)
    else:
        raise ValueError(style)


##


def substitute_params(args: lang.Args, values: ta.Mapping[ParamKey, ta.Any]) -> lang.Args:
    return lang.Args(
        *[values[a] for a in args.args],
        **{k: values[v] for k, v in args.kwargs.items()},
    )



"""
TODO:
 - select best debugger: ipdb, pudb,
"""
import dataclasses as dc
import functools
import traceback
import types
import typing as ta

from .. import check
from .. import lang


if ta.TYPE_CHECKING:
    import pdb  # noqa

    from . import pydevd

else:
    pdb = lang.proxy_import('pdb')

    pydevd = lang.proxy_import('.pydevd', __package__)


T = ta.TypeVar('T')
P = ta.ParamSpec('P')

ExceptionInfo: ta.TypeAlias = tuple[type[BaseException], BaseException, types.TracebackType]

PostMortemDebugger: ta.TypeAlias = ta.Callable[[ExceptionInfo], None]


##


def pdb_post_mortem_debugger(ei: ExceptionInfo) -> None:
    pdb.post_mortem(ei[2])


def get_post_mortem_debugger() -> PostMortemDebugger:
    if pydevd.is_present():
        return pydevd.debug_unhandled_exception

    else:
        check.not_none(pdb.post_mortem)
        return pdb_post_mortem_debugger


##


@dc.dataclass(kw_only=True)
class DebuggingOnException:
    silent: bool = False

    fn: PostMortemDebugger | None = None

    def __call__(self, fn: ta.Callable[P, T]) -> ta.Callable[P, T]:
        @functools.wraps(fn)
        def inner(*args, **kwargs):
            with self:
                return fn(*args, **kwargs)

        return inner

    def __enter__(self) -> ta.Self:
        if self.fn is None:
            self.fn = get_post_mortem_debugger()

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_val is None:
            return

        if not self.silent:
            traceback.print_exc()

        check.not_none(self.fn)((exc_type, exc_val, exc_tb))


debugging_on_exception = DebuggingOnException



import ctypes as ct
import sys


LINUX_PR_SET_PDEATHSIG = 1  # Second arg is a signal
LINUX_PR_GET_PDEATHSIG = 2  # Second arg is a ptr to return the signal


def set_process_deathsig(sig: int) -> bool:
    if sys.platform == 'linux':
        libc = ct.CDLL('libc.so.6')

        # int prctl(int option, unsigned long arg2, unsigned long arg3, unsigned long arg4, unsigned long arg5);
        libc.prctl.restype = ct.c_int
        libc.prctl.argtypes = [ct.c_int, ct.c_ulong, ct.c_ulong, ct.c_ulong, ct.c_ulong]

        libc.prctl(LINUX_PR_SET_PDEATHSIG, sig, 0, 0, 0, 0)

        return True

    else:
        return False



from .consts import (  # noqa
    DEFAULT_BUFFER_SIZE,
)

from .direct import (  # noqa
    DirectGenerator,

    BytesDirectGenerator,
    StrDirectGenerator,
)

from .readers import (  # noqa
    ReaderGenerator,
    BytesReaderGenerator,
    StrReaderGenerator,

    ExactReaderGenerator,
    BytesExactReaderGenerator,
    StrExactReaderGenerator,

    GeneratorReader,

    PrependableGeneratorReader,
    PrependableBytesGeneratorReader,
    PrependableStrGeneratorReader,
    prependable_bytes_generator_reader,
    prependable_str_generator_reader,

    BufferedGeneratorReader,
    BufferedBytesGeneratorReader,
    BufferedStrGeneratorReader,
    buffered_bytes_generator_reader,
    buffered_str_generator_reader,
)

from .stepped import (  # noqa
    SteppedGenerator,
    BytesSteppedGenerator,
    StrSteppedGenerator,
    BytesToStrSteppedGenerator,
    StrToBytesSteppedGenerator,

    SteppedReaderGenerator,
    BytesSteppedReaderGenerator,
    StrSteppedReaderGenerator,

    flatmap_stepped_generator,

    joined_bytes_stepped_generator,
    joined_str_stepped_generator,

    read_into_bytes_stepped_generator,
)



import typing as ta

from .. import check
from .frozen import FrozenDict
from .frozen import FrozenList


T = ta.TypeVar('T')
T2 = ta.TypeVar('T2')
K = ta.TypeVar('K')
K2 = ta.TypeVar('K2')
V = ta.TypeVar('V')
V2 = ta.TypeVar('V2')

_map = map


def _unpack_fn(fn):
    if isinstance(fn, tuple):
        return check.of_isinstance(fn)
    elif isinstance(fn, type) or callable(fn):
        return fn
    else:
        raise TypeError(fn)


# region seq


def seq(
        it: ta.Iterable[T],
) -> ta.Sequence[T]:
    if isinstance(it, str):
        raise TypeError(it)
    elif isinstance(it, FrozenList):
        return it
    else:
        return FrozenList(it)


def opt_seq(
        it: ta.Iterable[T] | None,
) -> ta.Sequence[T] | None:
    if it is None:
        return None
    else:
        return seq(it)


def seq_or_none(
        it: ta.Iterable[T] | None,
) -> ta.Sequence[T] | None:
    ret = opt_seq(it)
    if ret:
        return ret
    else:
        return None


def seq_of(
        fn: ta.Callable[[T], T2] | tuple,
) -> ta.Callable[[ta.Iterable[T]], ta.Sequence[T2]]:
    def inner(it):
        return seq(fn(e) for e in it)  # type: ignore

    fn = _unpack_fn(fn)
    return inner


def opt_seq_of(
        fn: ta.Callable[[T], T2] | tuple,
) -> ta.Callable[[ta.Iterable[T] | None], ta.Sequence[T2] | None]:
    def inner(it):
        if it is None:
            return None
        else:
            return seq(fn(e) for e in it)  # type: ignore

    fn = _unpack_fn(fn)
    return inner


def seq_of_or_none(
        fn: ta.Callable[[T], T2] | tuple,
) -> ta.Callable[[ta.Iterable[T] | None], ta.Sequence[T2] | None]:
    def inner(it):
        if it is None:
            return None
        else:
            ret = seq(fn(e) for e in it)  # type: ignore
            if ret:
                return ret
            else:
                return None

    fn = _unpack_fn(fn)
    return inner


# endregion


# region abs_set


def abs_set(
        it: ta.Iterable[T],
) -> ta.AbstractSet[T]:
    if isinstance(it, str):
        raise TypeError(it)
    elif isinstance(it, frozenset):
        return it
    else:
        return frozenset(it)


def opt_abs_set(
        it: ta.Iterable[T] | None,
) -> ta.AbstractSet[T] | None:
    if it is None:
        return None
    else:
        return abs_set(it)


def abs_set_or_none(
        it: ta.Iterable[T] | None,
) -> ta.AbstractSet[T] | None:
    ret = opt_abs_set(it)
    if ret:
        return ret
    else:
        return None


def abs_set_of(
        fn: ta.Callable[[T], T2] | tuple,
) -> ta.Callable[[ta.Iterable[T]], ta.AbstractSet[T2]]:
    def inner(it):
        return abs_set(fn(e) for e in it)  # type: ignore

    fn = _unpack_fn(fn)
    return inner


def opt_abs_set_of(
        fn: ta.Callable[[T], T2] | tuple,
) -> ta.Callable[[ta.Iterable[T] | None], ta.AbstractSet[T2] | None]:
    def inner(it):
        if it is None:
            return None
        else:
            return abs_set(fn(e) for e in it)  # type: ignore

    fn = _unpack_fn(fn)
    return inner


def abs_set_of_or_none(
        fn: ta.Callable[[T], T2] | tuple,
) -> ta.Callable[[ta.Iterable[T] | None], ta.AbstractSet[T2] | None]:
    def inner(it):
        if it is None:
            return None
        else:
            ret = abs_set(fn(e) for e in it)  # type: ignore
            if ret:
                return ret
            else:
                return None

    fn = _unpack_fn(fn)
    return inner


# endregion


# region frozenset


def frozenset_(
        it: ta.Iterable[T],
) -> frozenset[T]:
    if isinstance(it, str):
        raise TypeError(it)
    elif isinstance(it, frozenset):
        return it
    else:
        return frozenset(it)


def opt_frozenset(
        it: ta.Iterable[T] | None,
) -> frozenset[T] | None:
    if it is None:
        return None
    else:
        return frozenset_(it)


def frozenset_or_none(
        it: ta.Iterable[T] | None,
) -> frozenset[T] | None:
    ret = opt_frozenset(it)
    if ret:
        return ret
    else:
        return None


def frozenset_of(
        fn: ta.Callable[[T], T2] | tuple,
) -> ta.Callable[[ta.Iterable[T]], frozenset[T2]]:
    def inner(it):
        return frozenset_(fn(e) for e in it)  # type: ignore

    fn = _unpack_fn(fn)
    return inner


def opt_frozenset_of(
        fn: ta.Callable[[T], T2] | tuple,
) -> ta.Callable[[ta.Iterable[T] | None], frozenset[T2] | None]:
    def inner(it):
        if it is None:
            return None
        else:
            return frozenset_(fn(e) for e in it)  # type: ignore

    fn = _unpack_fn(fn)
    return inner


def frozenset_of_or_none(
        fn: ta.Callable[[T], T2] | tuple,
) -> ta.Callable[[ta.Iterable[T] | None], frozenset[T2] | None]:
    def inner(it):
        if it is None:
            return None
        else:
            ret = frozenset_(fn(e) for e in it)  # type: ignore
            if ret:
                return ret
            else:
                return None

    fn = _unpack_fn(fn)
    return inner


# endregion


# region map


def map(  # noqa
        src: ta.Mapping[K, V] | ta.Iterable[tuple[K, V]],
) -> ta.Mapping[K, V]:
    return FrozenDict(src)


def opt_map(
        src: ta.Mapping[K, V] | ta.Iterable[tuple[K, V]] | None,
) -> ta.Mapping[K, V] | None:
    if src is None:
        return None
    else:
        return map(src)


def map_or_none(
        src: ta.Mapping[K, V] | ta.Iterable[tuple[K, V]] | None,
) -> ta.Mapping[K, V] | None:
    ret = opt_map(src)
    if ret:
        return ret
    else:
        return None


def map_of(
        key_fn: ta.Callable[[K], K2] | tuple,
        value_fn: ta.Callable[[V], V2] | tuple,
) -> ta.Callable[
    [ta.Mapping[K, V] | ta.Iterable[tuple[K, V]]],
    ta.Mapping[K2, V2],
]:
    def inner(src):
        return map((key_fn(k), value_fn(v)) for k, v in dict(src).items())  # type: ignore

    key_fn = _unpack_fn(key_fn)
    value_fn = _unpack_fn(value_fn)
    return inner


def opt_map_of(
        key_fn: ta.Callable[[K], K2] | tuple,
        value_fn: ta.Callable[[V], V2] | tuple,
) -> ta.Callable[
    [ta.Mapping[K, V] | ta.Iterable[tuple[K, V]] | None],
    ta.Mapping[K2, V2] | None,
]:
    def inner(src):
        if src is None:
            return None
        else:
            return map((key_fn(k), value_fn(v)) for k, v in dict(src).items())  # type: ignore

    key_fn = _unpack_fn(key_fn)
    value_fn = _unpack_fn(value_fn)
    return inner


def map_of_or_none(
        key_fn: ta.Callable[[K], K2] | tuple,
        value_fn: ta.Callable[[V], V2] | tuple,
) -> ta.Callable[
    [ta.Mapping[K, V] | ta.Iterable[tuple[K, V]] | None],
    ta.Mapping[K2, V2] | None,
]:
    def inner(src):
        if src is None:
            return None
        else:
            ret = map((key_fn(k), value_fn(v)) for k, v in dict(src).items())  # type: ignore
            if ret:
                return ret
            else:
                return None

    key_fn = _unpack_fn(key_fn)
    value_fn = _unpack_fn(value_fn)
    return inner


# endregion



# ruff: noqa: UP006 UP007
import asyncio.base_subprocess
import asyncio.subprocess
import contextlib
import functools
import logging
import subprocess
import sys
import typing as ta

from ..check import check_equal
from ..check import check_isinstance
from ..check import check_not_none
from ..check import check_single
from ..logs import log
from ..subprocesses import DEFAULT_SUBPROCESS_TRY_EXCEPTIONS
from ..subprocesses import prepare_subprocess_invocation
from ..subprocesses import subprocess_common_context
from .asyncio import asyncio_maybe_timeout


T = ta.TypeVar('T')


##


@contextlib.asynccontextmanager
async def asyncio_subprocess_popen(
        *cmd: str,
        shell: bool = False,
        timeout: ta.Optional[float] = None,
        **kwargs: ta.Any,
) -> ta.AsyncGenerator[asyncio.subprocess.Process, None]:
    fac: ta.Any
    if shell:
        fac = functools.partial(
            asyncio.create_subprocess_shell,
            check_single(cmd),
        )
    else:
        fac = functools.partial(
            asyncio.create_subprocess_exec,
            *cmd,
        )

    with subprocess_common_context(
            *cmd,
            shell=shell,
            timeout=timeout,
            **kwargs,
    ):
        proc: asyncio.subprocess.Process
        proc = await fac(**kwargs)
        try:
            yield proc

        finally:
            await asyncio_maybe_timeout(proc.wait(), timeout)


##


class AsyncioProcessCommunicator:
    def __init__(
            self,
            proc: asyncio.subprocess.Process,
            loop: ta.Optional[ta.Any] = None,
    ) -> None:
        super().__init__()

        if loop is None:
            loop = asyncio.get_running_loop()

        self._proc = proc
        self._loop = loop

        self._transport: asyncio.base_subprocess.BaseSubprocessTransport = check_isinstance(
            proc._transport,  # type: ignore  # noqa
            asyncio.base_subprocess.BaseSubprocessTransport,
        )

    @property
    def _debug(self) -> bool:
        return self._loop.get_debug()

    async def _feed_stdin(self, input: bytes) -> None:  # noqa
        stdin = check_not_none(self._proc.stdin)
        try:
            if input is not None:
                stdin.write(input)
                if self._debug:
                    log.debug('%r communicate: feed stdin (%s bytes)', self, len(input))

            await stdin.drain()

        except (BrokenPipeError, ConnectionResetError) as exc:
            # communicate() ignores BrokenPipeError and ConnectionResetError. write() and drain() can raise these
            # exceptions.
            if self._debug:
                log.debug('%r communicate: stdin got %r', self, exc)

        if self._debug:
            log.debug('%r communicate: close stdin', self)

        stdin.close()

    async def _noop(self) -> None:
        return None

    async def _read_stream(self, fd: int) -> bytes:
        transport: ta.Any = check_not_none(self._transport.get_pipe_transport(fd))

        if fd == 2:
            stream = check_not_none(self._proc.stderr)
        else:
            check_equal(fd, 1)
            stream = check_not_none(self._proc.stdout)

        if self._debug:
            name = 'stdout' if fd == 1 else 'stderr'
            log.debug('%r communicate: read %s', self, name)

        output = await stream.read()

        if self._debug:
            name = 'stdout' if fd == 1 else 'stderr'
            log.debug('%r communicate: close %s', self, name)

        transport.close()

        return output

    class Communication(ta.NamedTuple):
        stdout: ta.Optional[bytes]
        stderr: ta.Optional[bytes]

    async def _communicate(
            self,
            input: ta.Any = None,  # noqa
    ) -> Communication:
        stdin_fut: ta.Any
        if self._proc.stdin is not None:
            stdin_fut = self._feed_stdin(input)
        else:
            stdin_fut = self._noop()

        stdout_fut: ta.Any
        if self._proc.stdout is not None:
            stdout_fut = self._read_stream(1)
        else:
            stdout_fut = self._noop()

        stderr_fut: ta.Any
        if self._proc.stderr is not None:
            stderr_fut = self._read_stream(2)
        else:
            stderr_fut = self._noop()

        stdin_res, stdout_res, stderr_res = await asyncio.gather(stdin_fut, stdout_fut, stderr_fut)

        await self._proc.wait()

        return AsyncioProcessCommunicator.Communication(stdout_res, stderr_res)

    async def communicate(
            self,
            input: ta.Any = None,  # noqa
            timeout: ta.Optional[float] = None,
    ) -> Communication:
        return await asyncio_maybe_timeout(self._communicate(input), timeout)


async def asyncio_subprocess_communicate(
        proc: asyncio.subprocess.Process,
        input: ta.Any = None,  # noqa
        timeout: ta.Optional[float] = None,
) -> ta.Tuple[ta.Optional[bytes], ta.Optional[bytes]]:
    return await AsyncioProcessCommunicator(proc).communicate(input, timeout)  # noqa


##


async def _asyncio_subprocess_check_run(
        *args: str,
        input: ta.Any = None,  # noqa
        timeout: ta.Optional[float] = None,
        **kwargs: ta.Any,
) -> ta.Tuple[ta.Optional[bytes], ta.Optional[bytes]]:
    args, kwargs = prepare_subprocess_invocation(*args, **kwargs)

    proc: asyncio.subprocess.Process
    async with asyncio_subprocess_popen(*args, **kwargs) as proc:
        stdout, stderr = await asyncio_subprocess_communicate(proc, input, timeout)

    if proc.returncode:
        raise subprocess.CalledProcessError(
            proc.returncode,
            args,
            output=stdout,
            stderr=stderr,
        )

    return stdout, stderr


async def asyncio_subprocess_check_call(
        *args: str,
        stdout: ta.Any = sys.stderr,
        input: ta.Any = None,  # noqa
        timeout: ta.Optional[float] = None,
        **kwargs: ta.Any,
) -> None:
    _, _ = await _asyncio_subprocess_check_run(
        *args,
        stdout=stdout,
        input=input,
        timeout=timeout,
        **kwargs,
    )


async def asyncio_subprocess_check_output(
        *args: str,
        input: ta.Any = None,  # noqa
        timeout: ta.Optional[float] = None,
        **kwargs: ta.Any,
) -> bytes:
    stdout, stderr = await _asyncio_subprocess_check_run(
        *args,
        stdout=asyncio.subprocess.PIPE,
        input=input,
        timeout=timeout,
        **kwargs,
    )

    return check_not_none(stdout)


async def asyncio_subprocess_check_output_str(*args: str, **kwargs: ta.Any) -> str:
    return (await asyncio_subprocess_check_output(*args, **kwargs)).decode().strip()


##


async def _asyncio_subprocess_try_run(
        fn: ta.Callable[..., ta.Awaitable[T]],
        *args: ta.Any,
        try_exceptions: ta.Tuple[ta.Type[Exception], ...] = DEFAULT_SUBPROCESS_TRY_EXCEPTIONS,
        **kwargs: ta.Any,
) -> ta.Union[T, Exception]:
    try:
        return await fn(*args, **kwargs)
    except try_exceptions as e:  # noqa
        if log.isEnabledFor(logging.DEBUG):
            log.exception('command failed')
        return e


async def asyncio_subprocess_try_call(
        *args: str,
        try_exceptions: ta.Tuple[ta.Type[Exception], ...] = DEFAULT_SUBPROCESS_TRY_EXCEPTIONS,
        **kwargs: ta.Any,
) -> bool:
    if isinstance(await _asyncio_subprocess_try_run(
            asyncio_subprocess_check_call,
            *args,
            try_exceptions=try_exceptions,
            **kwargs,
    ), Exception):
        return False
    else:
        return True


async def asyncio_subprocess_try_output(
        *args: str,
        try_exceptions: ta.Tuple[ta.Type[Exception], ...] = DEFAULT_SUBPROCESS_TRY_EXCEPTIONS,
        **kwargs: ta.Any,
) -> ta.Optional[bytes]:
    if isinstance(ret := await _asyncio_subprocess_try_run(
            asyncio_subprocess_check_output,
            *args,
            try_exceptions=try_exceptions,
            **kwargs,
    ), Exception):
        return None
    else:
        return ret


async def asyncio_subprocess_try_output_str(*args: str, **kwargs: ta.Any) -> ta.Optional[str]:
    out = await asyncio_subprocess_try_output(*args, **kwargs)
    return out.decode().strip() if out is not None else None



# ruff: noqa: UP006 UP007
import socket
import socketserver
import typing as ta

from omlish.lite.check import check_not_none

from .socket import SocketAddress
from .socket import SocketHandlerFactory


##


class SocketServerBaseRequestHandler_:  # noqa
    request: socket.socket
    client_address: SocketAddress
    server: socketserver.TCPServer


class SocketServerStreamRequestHandler_(SocketServerBaseRequestHandler_):  # noqa
    rbufsize: int
    wbufsize: int

    timeout: ta.Optional[float]

    disable_nagle_algorithm: bool

    connection: socket.socket
    rfile: ta.BinaryIO
    wfile: ta.BinaryIO


##


class SocketHandlerSocketServerStreamRequestHandler(  # type: ignore[misc]
    socketserver.StreamRequestHandler,
    SocketServerStreamRequestHandler_,
):
    socket_handler_factory: ta.Optional[SocketHandlerFactory] = None

    def __init__(
            self,
            request: socket.socket,
            client_address: SocketAddress,
            server: socketserver.TCPServer,
            *,
            socket_handler_factory: ta.Optional[SocketHandlerFactory] = None,
    ) -> None:
        if socket_handler_factory is not None:
            self.socket_handler_factory = socket_handler_factory

        super().__init__(
            request,
            client_address,
            server,
        )

    def handle(self) -> None:
        target = check_not_none(self.socket_handler_factory)(
            self.client_address,
            self.rfile,  # type: ignore[arg-type]
            self.wfile,  # type: ignore[arg-type]
        )
        target.handle()



import dataclasses as dc
import typing as ta

from ... import lang
from .base import Compression


if ta.TYPE_CHECKING:
    import zstandard
else:
    zstandard = lang.proxy_import('zstandard')


@dc.dataclass(frozen=True, kw_only=True)
class ZstdCompression(Compression):
    level: int | None = None

    max_output_size: int = 0

    def compress(self, d: bytes) -> bytes:
        return zstandard.compress(
            d,
            **(dict(level=self.level) if self.level is not None else {}),
        )

    def decompress(self, d: bytes) -> bytes:
        return zstandard.decompress(
            d,
            max_output_size=self.max_output_size,
        )



import dataclasses as dc
import typing as ta

from ... import lang
from .base import Compression


if ta.TYPE_CHECKING:
    import snappy
else:
    snappy = lang.proxy_import('snappy')


@dc.dataclass(frozen=True)
class SnappyCompression(Compression):
    def compress(self, d: bytes) -> bytes:
        return snappy.compress(d)

    def decompress(self, d: bytes) -> bytes:
        return snappy.decompress(d)



import typing as ta


O = ta.TypeVar('O')
I = ta.TypeVar('I')
R = ta.TypeVar('R')


# Direct generators yield outputs 1:1 with inputs.
DirectGenerator: ta.TypeAlias = ta.Generator[O, I, R]

BytesDirectGenerator: ta.TypeAlias = DirectGenerator[bytes, bytes, R]
StrDirectGenerator: ta.TypeAlias = DirectGenerator[str, str, R]



import abc
import itertools
import typing as ta

from .. import check
from .. import lang


K = ta.TypeVar('K')
V = ta.TypeVar('V')
StrMap = ta.Mapping[str, ta.Any]


class _MISSING(lang.Marker):
    pass


class Flattening:

    DEFAULT_DELIMITER = '.'
    DEFAULT_INDEX_OPEN = '('
    DEFAULT_INDEX_CLOSE = ')'

    def __init__(
            self,
            *,
            delimiter=DEFAULT_DELIMITER,
            index_open=DEFAULT_INDEX_OPEN,
            index_close=DEFAULT_INDEX_CLOSE,
    ) -> None:
        super().__init__()

        self._delimiter = check.not_empty(delimiter)
        self._index_open = check.not_empty(index_open)
        self._index_close = check.not_empty(index_close)

    def flatten(self, unflattened: StrMap) -> StrMap:
        def rec(prefix: list[str], value: ta.Any) -> None:
            if isinstance(value, dict):
                for k, v in value.items():
                    rec([*prefix, k], v)
            elif isinstance(value, list):
                check.not_empty(prefix)
                for i, v in enumerate(value):
                    rec(prefix[:-1] + [f'{prefix[-1]}{self._index_open}{i}{self._index_close}'], v)
            else:
                k = self._delimiter.join(prefix)
                if k in ret:
                    raise KeyError(k)
                ret[k] = value

        ret: dict[str, ta.Any] = {}
        rec([], unflattened)
        return ret

    class UnflattenNode(lang.Abstract, ta.Generic[K]):

        @abc.abstractmethod
        def get(self, key: K) -> ta.Any:
            raise NotImplementedError

        @abc.abstractmethod
        def put(self, key: K, value: ta.Any) -> None:
            raise NotImplementedError

        def setdefault(self, key: K, supplier: ta.Callable[[], V]) -> V:
            ret = self.get(key)
            if ret is _MISSING:
                ret = supplier()
                self.put(key, ret)
            return ret

        @abc.abstractmethod
        def build(self) -> ta.Any:
            raise NotImplementedError

        @staticmethod
        def maybe_build(value: ta.Any) -> ta.Any:
            check.not_none(value)
            return value.build() if isinstance(value, Flattening.UnflattenNode) else value

    class UnflattenDict(UnflattenNode[str]):

        def __init__(self) -> None:
            super().__init__()

            self._dict: dict[str, ta.Any] = {}

        def get(self, key: str) -> ta.Any:
            return self._dict.get(key, _MISSING)

        def put(self, key: str, value: ta.Any) -> None:
            check.arg(key not in self._dict)
            self._dict[key] = value

        def build(self) -> ta.Any:
            return {k: Flattening.UnflattenNode.maybe_build(v) for k, v in self._dict.items()}

    class UnflattenList(UnflattenNode[int]):

        def __init__(self) -> None:
            super().__init__()

            self._list: list[ta.Any] = []

        def get(self, key: int) -> ta.Any:
            check.arg(key >= 0)
            return self._list[key] if key < len(self._list) else _MISSING

        def put(self, key: int, value: ta.Any) -> None:
            check.arg(key >= 0)
            if key >= len(self._list):
                self._list.extend([_MISSING] * (key - len(self._list) + 1))
            check.arg(self._list[key] is _MISSING)
            self._list[key] = value

        def build(self) -> ta.Any:
            return [Flattening.UnflattenNode.maybe_build(e) for e in self._list]

    def unflatten(self, flattened: StrMap) -> StrMap:
        root = Flattening.UnflattenDict()

        def split_keys(fkey: str) -> ta.Iterable[str | int]:
            for part in fkey.split(self._delimiter):
                if self._index_open in part:
                    check.state(part.endswith(self._index_close))
                    pos = part.index(self._index_open)
                    yield part[:pos]
                    for p in part[pos + len(self._index_open):-len(self._index_close)] \
                            .split(self._index_close + self._index_open):
                        yield int(p)
                else:
                    check.state(self._index_close not in part)
                    yield part

        for fk, v in flattened.items():
            node: Flattening.UnflattenNode = root
            fks = list(split_keys(fk))
            for key, nkey in itertools.pairwise(fks):
                if isinstance(nkey, str):
                    node = node.setdefault(key, Flattening.UnflattenDict)
                elif isinstance(nkey, int):
                    node = node.setdefault(key, Flattening.UnflattenList)
                else:
                    raise TypeError(key)
            node.put(fks[-1], v)

        return root.build()



import dataclasses as dc
import typing as ta

from ... import check
from ... import lang
from ..generators import BytesSteppedGenerator
from .base import Compression
from .base import IncrementalCompression


if ta.TYPE_CHECKING:
    import lz4.frame as lz4_frame
else:
    lz4_frame = lang.proxy_import('lz4.frame')


@dc.dataclass(frozen=True, kw_only=True)
class Lz4Compression(Compression, IncrementalCompression):
    level: int = 0

    block_size: int = 0
    block_linked: bool = True
    block_checksum: bool = False
    content_checksum: bool = False
    store_size: bool = True
    auto_flush: bool = False

    def compress(self, d: bytes) -> bytes:
        return lz4_frame.compress(
            d,
            compression_level=self.level,
            block_size=self.block_size,
            content_checksum=self.content_checksum,
            block_linked=self.block_linked,
            store_size=self.store_size,
        )

    def decompress(self, d: bytes) -> bytes:
        return lz4_frame.decompress(
            d,
        )

    @lang.autostart
    def compress_incremental(self) -> BytesSteppedGenerator[None]:
        with lz4_frame.LZ4FrameCompressor(
                compression_level=self.level,
                block_size=self.block_size,
                block_linked=self.block_linked,
                block_checksum=self.block_checksum,
                content_checksum=self.content_checksum,
                auto_flush=self.auto_flush,
        ) as compressor:
            started = False
            while True:
                i = check.isinstance((yield None), bytes)
                if not started:
                    yield compressor.begin()
                    started = True
                if not i:
                    yield compressor.flush()
                    yield b''
                    return
                if (o := compressor.compress(i)):
                    yield o

    @lang.autostart
    def decompress_incremental(self) -> BytesSteppedGenerator[None]:
        # lz4 lib does internal buffering so this is simply a BytesSteppedGenerator not a BytesSteppedReaderGenerator as
        # it only yields None, accepting any number of bytes at a time.
        with lz4_frame.LZ4FrameDecompressor() as decompressor:
            while True:
                i = check.isinstance((yield None), bytes)
                if not i:
                    yield b''
                    return
                if (o := decompressor.decompress(i)):
                    yield o



from .items import (  # noqa
    Attrs,
    Cell,
    Edge,
    Graph,
    Id,
    Item,
    Node,
    Raw,
    RawStmt,
    Row,
    Stmt,
    Table,
    Text,
)

from .make import (  # noqa
    make_simple,
)

from .rendering import (  # noqa
    Renderer,
    open_dot,
    render,
)

from .utils import (  # noqa
    Color,
    escape,
    gen_rainbow,
)



import functools
import typing as ta


T = ta.TypeVar('T')

CallableT = ta.TypeVar('CallableT', bound=ta.Callable)


##


class _AbstractCachedNullary:
    def __init__(self, fn):
        super().__init__()
        self._fn = fn
        self._value = self._missing = object()
        functools.update_wrapper(self, fn)

    def __call__(self, *args, **kwargs):  # noqa
        raise TypeError

    def __get__(self, instance, owner):  # noqa
        bound = instance.__dict__[self._fn.__name__] = self.__class__(self._fn.__get__(instance, owner))
        return bound


##


class _CachedNullary(_AbstractCachedNullary):
    def __call__(self, *args, **kwargs):  # noqa
        if self._value is self._missing:
            self._value = self._fn()
        return self._value


def cached_nullary(fn):  # ta.Callable[..., T]) -> ta.Callable[..., T]:
    return _CachedNullary(fn)


def static_init(fn: CallableT) -> CallableT:
    fn = cached_nullary(fn)
    fn()
    return fn


##


class _AsyncCachedNullary(_AbstractCachedNullary):
    async def __call__(self, *args, **kwargs):
        if self._value is self._missing:
            self._value = await self._fn()
        return self._value


def async_cached_nullary(fn):  # ta.Callable[..., T]) -> ta.Callable[..., T]:
    return _AsyncCachedNullary(fn)



import enum
import typing as ta

from ... import check
from ... import dataclasses as dc
from ... import lang
from .exprs import CanExpr
from .exprs import Expr
from .exprs import ExprBuilder


##


class MultiKind(enum.Enum):
    AND = enum.auto()
    OR = enum.auto()


class Multi(Expr, lang.Final):
    k: MultiKind
    es: ta.Sequence[Expr] = dc.xfield(coerce=tuple, validate=lambda es: len(es) > 1)


class MultiBuilder(ExprBuilder):
    def multi(self, k: MultiKind, *es: CanExpr) -> Expr:
        check.not_empty(es)
        if len(es) == 1:
            return self.expr(es[0])
        else:
            return Multi(k, tuple(self.expr(e) for e in es))

    def and_(self, *es: CanExpr) -> Expr:
        return self.multi(MultiKind.AND, *es)

    def or_(self, *es: CanExpr) -> Expr:
        return self.multi(MultiKind.OR, *es)



import re
import sys


DOCKER_FOR_MAC_HOSTNAME = 'docker.for.mac.localhost'


_LIKELY_IN_DOCKER_PATTERN = re.compile(r'^overlay / .*/(docker|desktop-containerd)/')


def is_likely_in_docker() -> bool:
    if getattr(sys, 'platform') != 'linux':
        return False
    with open('/proc/mounts') as f:
        ls = f.readlines()
    return any(_LIKELY_IN_DOCKER_PATTERN.match(l) for l in ls)



import dataclasses as dc
import functools
import typing as ta

from ... import lang
from ..generators import BytesSteppedGenerator
from ..generators import BytesSteppedReaderGenerator
from .adapters import CompressorObjectIncrementalAdapter
from .adapters import DecompressorObjectIncrementalAdapter
from .base import Compression
from .base import IncrementalCompression


if ta.TYPE_CHECKING:
    import bz2
else:
    bz2 = lang.proxy_import('bz2')


@dc.dataclass(frozen=True, kw_only=True)
class Bz2Compression(Compression, IncrementalCompression):
    level: int = 9

    def compress(self, d: bytes) -> bytes:
        return bz2.compress(
            d,
            self.level,
        )

    def decompress(self, d: bytes) -> bytes:
        return bz2.decompress(
            d,
        )

    def compress_incremental(self) -> BytesSteppedGenerator[None]:
        return lang.nextgen(CompressorObjectIncrementalAdapter(
            functools.partial(
                bz2.BZ2Compressor,  # type: ignore
                self.level,
            ),
        )())

    def decompress_incremental(self) -> BytesSteppedReaderGenerator[None]:
        return DecompressorObjectIncrementalAdapter(
            bz2.BZ2Decompressor,  # type: ignore
            trailing_error=OSError,
        )()



"""
TODO:
 - reflecty generalized rewriter, obviously..
 - env vars
 - coalescing - {$FOO|$BAR|baz}
"""
import collections.abc
import typing as ta

from .. import dataclasses as dc
from .. import lang
from ..text import glyphsplit


T = ta.TypeVar('T')


class InterpolateStringsMetadata(lang.Marker):
    pass


@dc.field_modifier
def secret_or_key_field(f: dc.Field) -> dc.Field:
    return dc.update_field_metadata(f, {
        InterpolateStringsMetadata: True,
    })


@dc.field_modifier
def interpolate_field(f: dc.Field) -> dc.Field:
    return dc.update_field_metadata(f, {InterpolateStringsMetadata: True})


class StringRewriter:
    def __init__(self, fn: ta.Callable[[str], str]) -> None:
        super().__init__()
        self._fn = fn

    def __call__(self, v: T, *, _soft: bool = False) -> T:
        if v is None:
            return None  # type: ignore

        if dc.is_dataclass(v):
            kw = {}
            for f in dc.fields(v):
                fv = getattr(v, f.name)
                nfv = self(fv, _soft=not f.metadata.get(InterpolateStringsMetadata))
                if fv is not nfv:
                    kw[f.name] = nfv
            if not kw:
                return v  # type: ignore
            return dc.replace(v, **kw)

        if isinstance(v, str):
            if not _soft:
                v = self._fn(v)  # type: ignore
            return v  # type: ignore

        if isinstance(v, lang.BUILTIN_SCALAR_ITERABLE_TYPES):
            return v  # type: ignore

        if isinstance(v, collections.abc.Mapping):
            nm = []
            b = False
            for mk, mv in v.items():
                nk, nv = self(mk, _soft=_soft), self(mv, _soft=_soft)
                nm.append((nk, nv))
                b |= nk is not mk or nv is not mv
            if not b:
                return v  # type: ignore
            return v.__class__(nm)  # type: ignore

        if isinstance(v, (collections.abc.Sequence, collections.abc.Set)):
            nl = []
            b = False
            for le in v:
                ne = self(le, _soft=_soft)
                nl.append(ne)
                b |= ne is not le
            if not b:
                return v  # type: ignore
            return v.__class__(nl)  # type: ignore

        return v


def interpolate_strings(v: T, rpl: ta.Mapping[str, str]) -> T:
    def fn(v):
        if not v:
            return v
        sps = glyphsplit.split_braces(v)
        if len(sps) == 1 and isinstance(sps[0], str):
            return sps[0]
        return ''.join(rpl[p.s] if isinstance(p, glyphsplit.GlyphMatch) else p for p in sps)

    return StringRewriter(fn)(v)



# ruff: noqa: UP006 UP007
import socket
import typing as ta

from ...lite.check import check_isinstance
from ...lite.check import check_none
from ...lite.check import check_not_none
from ...lite.check import check_state
from ...lite.http.coroserver import CoroHttpServer
from ...lite.http.handlers import HttpHandler
from ...lite.socket import SocketAddress
from ..buffers import IncrementalWriteBuffer
from ..buffers import ReadableListBuffer
from .handlers import SocketFdioHandler


class CoroHttpServerConnectionFdioHandler(SocketFdioHandler):
    def __init__(
            self,
            addr: SocketAddress,
            sock: socket.socket,
            handler: HttpHandler,
            *,
            read_size: int = 0x10000,
            write_size: int = 0x10000,
            log_handler: ta.Optional[ta.Callable[[CoroHttpServer, CoroHttpServer.AnyLogIo], None]] = None,
    ) -> None:
        check_state(not sock.getblocking())

        super().__init__(addr, sock)

        self._handler = handler
        self._read_size = read_size
        self._write_size = write_size
        self._log_handler = log_handler

        self._read_buf = ReadableListBuffer()
        self._write_buf: IncrementalWriteBuffer | None = None

        self._coro_srv = CoroHttpServer(
            addr,
            handler=self._handler,
        )
        self._srv_coro: ta.Optional[ta.Generator[CoroHttpServer.Io, ta.Optional[bytes], None]] = self._coro_srv.coro_handle()  # noqa

        self._cur_io: CoroHttpServer.Io | None = None
        self._next_io()

    #

    def _next_io(self) -> None:  # noqa
        coro = check_not_none(self._srv_coro)

        d: bytes | None = None
        o = self._cur_io
        while True:
            if o is None:
                try:
                    if d is not None:
                        o = coro.send(d)
                        d = None
                    else:
                        o = next(coro)
                except StopIteration:
                    self.close()
                    o = None
                    break

            if isinstance(o, CoroHttpServer.AnyLogIo):
                if self._log_handler is not None:
                    self._log_handler(self._coro_srv, o)
                o = None

            elif isinstance(o, CoroHttpServer.ReadIo):
                if (d := self._read_buf.read(o.sz)) is None:
                    break
                o = None

            elif isinstance(o, CoroHttpServer.ReadLineIo):
                if (d := self._read_buf.read_until(b'\n')) is None:
                    break
                o = None

            elif isinstance(o, CoroHttpServer.WriteIo):
                check_none(self._write_buf)
                self._write_buf = IncrementalWriteBuffer(o.data, write_size=self._write_size)
                break

            else:
                raise TypeError(o)

        self._cur_io = o

    #

    def readable(self) -> bool:
        return True

    def writable(self) -> bool:
        return self._write_buf is not None

    #

    def on_readable(self) -> None:
        try:
            buf = check_not_none(self._sock).recv(self._read_size)
        except BlockingIOError:
            return
        except ConnectionResetError:
            self.close()
            return
        if not buf:
            self.close()
            return

        self._read_buf.feed(buf)

        if isinstance(self._cur_io, CoroHttpServer.AnyReadIo):
            self._next_io()

    def on_writable(self) -> None:
        check_isinstance(self._cur_io, CoroHttpServer.WriteIo)
        wb = check_not_none(self._write_buf)
        while wb.rem > 0:
            def send(d: bytes) -> int:
                try:
                    return check_not_none(self._sock).send(d)
                except ConnectionResetError:
                    self.close()
                    return 0
                except BlockingIOError:
                    return 0
            if not wb.write(send):
                break

        if wb.rem < 1:
            self._write_buf = None
            self._cur_io = None
            self._next_io()
